/*
  eFlexPwm SPWM

  These are 2 pairs of SPWM signals (Sinusoidal Pulse Width Modulation)
  Each signal pair corresponds to the PWMA output and the PWMB (A's complement)
  output of an eFlexPWM submodule.
  In this example, sub-modules 0 and 2 of PWM2 are used. On the Teensy 4.1 board,
  this corresponds to the following pins:
  | PWM | SubModule | Pol | Teensy | Native |
  |-----|-----------|-----|--------|--------|
  | 2   | 0         | A   | 4      | EMC_06 |
  | 2   | 0         | B   | 33     | EMC_07 |
  | 2   | 2         | A   | 6      | B0_10  |
  | 2   | 2         | B   | 9      | B0_11  |
  The PWM frequency is set by pwmFreq. Defaults to 500kHz

  This program works under interrupt and takes advantage of the i.MX RT1062 computing power.
  The Teensy's LED is lit during the interrupt routine in order to be able to measure its execution time.
/*

/* Core */
#include <Arduino.h>
#include <arm_math.h>
#include <math.h>
#include <util/atomic.h>
#include <eFlexPwm.h>
#define TIMEOUT 3000;

/* OLED screen */
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <Fonts/FreeSans9pt7b.h>
#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64
#define SCREEN_ADDRESS 0x3C
#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

using namespace eFlex;

const float32_t DeadTimeNs = 0.0;    // deadtime in nanoseconds
const uint32_t StepDelayMilliseconds = 1000;      // Increments sineFreq by SineFreqStep every StepDelay ms.
const bool ShouldOutputWaveValues = true;
const float32_t MaxDutyCycle = 32768;
const float32_t MinDutyCycle = -32768;
const int BlueLedPin = 31;
const int RedLedPin = 32;
const int SineFrequencyPotPin = 14; // A0
const int PwmFrequencyPotPin = 15; // A1
const int PhaseShiftPotPin = 16; // A2
const float32_t LedTolerance = 0.174532925f;

float32_t pwmFreq = 500000.0; // PWM frequency in hz
float32_t sineFreq = 50.0; // Sine Frequency in Hz
float32_t phaseShift = PI; // 180°
float32_t amplitudeShift = 0.2;
float32_t sinWavePeak;
float32_t sinWaveTrough;
float32_t radiansToDegreesMultiplier;
float32_t degreesToRadiansMultiplier;
float32_t degrees90InRadians;
float32_t degrees180InRadians;
float32_t degrees270InRadians;
float32_t degrees360InRadians;

// Log
const int LogSize = 5;
String logs[] = {
    "Line 1",
    "Line 2",
    "Line 3",
    "Line 4",
    "Line 5"
};

// Interrupt Input Variables
volatile uint32_t vElapsedIntervals;
volatile float32_t vUpdateIntervalPeriod;

// My eFlexPWM submodules (Hardware > PWM2: SM[0], SM[2])
SubModule Sm20 (4, 33);
SubModule Sm22 (6, 9);

/*
  all the sub-modules are part of the same timer: PWM2,
  Tm2 simplifies access to the functions that concern all the sub-modules
*/
Timer &Tm2 = Sm20.timer();

void setup() {
  Serial.begin(9600);

  // initialize LED digital pin as an output
  pinMode (LED_BUILTIN, OUTPUT);
  pinMode (BlueLedPin, OUTPUT);
  pinMode (RedLedPin, OUTPUT);
  pinMode (PwmFrequencyPotPin, INPUT);
  pinMode (SineFrequencyPotPin, INPUT);
  pinMode (PhaseShiftPotPin, INPUT);

  digitalWrite (LED_BUILTIN, HIGH);  
  digitalWrite (BlueLedPin, HIGH);
  digitalWrite (RedLedPin, HIGH);

  // SSD1306_SWITCHCAPVCC = generate display voltage from 3.3V internally
  if(!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));
    for(;;); // Don't proceed, loop forever
  }

  display.display();
  delay(2000); // Pause for 2 seconds

  // Clear the buffer
  display.clearDisplay();
  //display.setFont(&FreeSans9pt7b);
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setTextWrap(false);

  writeLog("eFlexPwm SPWM");

  preComputeValues();

  // start of PWM configuration -----------------------------------------------
  /* Submodule configuration, default values are:

    config->debugModeEnabled = false;
    config->enableWait = false;
    config->reloadSelect = kPWM_LocalReload;
    config->clockSource = kPWM_BusClock;
    config->prescale = kPWM_Prescale_Divide_1;
    config->initializationControl = kPWM_Initialize_LocalSync;
    config->forceTrigger = kPWM_Force_Local;
    config->reloadFrequency = kPWM_LoadEveryOportunity;
    config->reloadLogic = kPWM_ReloadImmediate;
    config->pairOperation = kPWM_Independent;
    config->pwmFreqHz = 5000;
    config->mode = kPWM_CenterAligned;
  */
  Config cfg;

  /* Use full cycle reload */
  cfg.setReloadLogic (kPWM_ReloadPwmFullCycle);
  /* PWM A & PWM B form a complementary PWM pair */
  cfg.setPairOperation (kPWM_ComplementaryPwmA);
  cfg.setPwmFreqHz (pwmFreq);

  /* Initialize submodule 0 */
  if (Sm20.configure (cfg) != true) {
    writeLog ("Submodule 0 initialization failed");
    exit (EXIT_FAILURE);
  }

  /* Initialize submodule 2, make it use same counter clock as submodule 0. */
  cfg.setClockSource (kPWM_Submodule0Clock);
  cfg.setPrescale (kPWM_Prescale_Divide_1);
  cfg.setInitializationControl (kPWM_Initialize_MasterSync);

  if (Sm22.configure (cfg) != true) {
    writeLog ("Submodule 2 initialization failed");
    exit (EXIT_FAILURE);
  }

  /*
    Set deadtime count, we set this to about 1000ns for all submodules pins
    You can also, modify it for all the pins of a sub-module, eg. :

      Sm20.setupDeadtime (deadTimeVal);

    or even for a particular pin, eg. :

      Sm20.setupDeadtime (deadTimeVal, ChanA);
  */
  uint16_t deadTimeVal = ( (uint64_t) Tm2.srcClockHz() * DeadTimeNs) / 1000000000;
  Tm2.setupDeadtime (deadTimeVal);

  // Start all submodule
  if (Tm2.begin() != true) {
    writeLog ("Failed to start submodules");
    exit (EXIT_FAILURE);
  }

  // synchronize registers and start all submodules
  if (Tm2.begin() != true) {
    writeLog ("Failed to start submodules");
    exit (EXIT_FAILURE);
  }
  else {

    writeLog ("Submodules successfuly started");
    // Sm20.printRegs(); // to see the values of the VALx registers
  }

  calculateUpdateIntervalPeriod();

  // Enable counter overflow interrupt (VAL1)
  Sm20.enableInterrupts (kPWM_CompareVal1InterruptEnable);
  attachInterruptVector (IRQ_FLEXPWM2_0, &IsrOverflowSm20);
  NVIC_ENABLE_IRQ (IRQ_FLEXPWM2_0);

  // end of PWM setup
  digitalWrite (LED_BUILTIN, LOW);
}

void loop() {
  delay (StepDelayMilliseconds);

  // Analogue potentiometers are not very accurate. Replace with digital input.
  int pwmFreqVal = analogRead(PwmFrequencyPotPin);
  pwmFreqVal = map(pwmFreqVal, 0, 1023, 1000, 50000);
  //pwmFreq = pwmFreqVal;

  int sineFreqVal = analogRead(SineFrequencyPotPin);
  sineFreqVal = map(sineFreqVal, 0, 1023, 1, 100);
  //sineFreq = sineFreqVal;
  sineFreq = 10;

  int phaseShiftVal = analogRead(PhaseShiftPotPin);
  phaseShiftVal = map(phaseShiftVal, 0, 1023, 0, degrees360InRadians);
  //phaseShift = phaseShiftVal;

  calculateUpdateIntervalPeriod();

  char strBuf[50];
  sprintf(strBuf, "Sine:%.1fHz", sineFreq);
  writeLog(strBuf);

  // Clear buffer
  memset (strBuf, 0, sizeof (strBuf));

  sprintf(strBuf, "PWM:%.1fHz", pwmFreq);
  writeLog(strBuf);

  // Clear buffer
  memset (strBuf, 0, sizeof (strBuf));

  sprintf(strBuf, "Phase:%.1f deg", radiansToDegrees(phaseShift));
  writeLog(strBuf);
/*
  // Clear buffer
  memset (strBuf, 0, sizeof (strBuf));

  sprintf(strBuf, "Amplitude:%.1f", amplitudeShift);
  writeLog(strBuf);
*/
  writeLog("Wave 1 peak:90 deg");

  // Clear buffer
  memset (strBuf, 0, sizeof (strBuf));

  sprintf(strBuf, "Wave 2 peak:%.1f deg", 90 + radiansToDegrees(phaseShift));
  writeLog(strBuf);
}

void calculateUpdateIntervalPeriod() {

  // Example
  // Sine Frequency = 50Hz
  // PWM Frequency = 10,000Hz
  // 2π = 360°
  // Update interval = 2π(50/10000)
  // Update interval = 2π(0.005)
  // Update interval = every 0.031415926535898 rads

  float32_t s = 2.0 * PI * (sineFreq / pwmFreq);
  ATOMIC_BLOCK (ATOMIC_RESTORESTATE) {
    vUpdateIntervalPeriod = s;
  }
}

// Triangle wave:
// y(x)=(2a/π)arcsin(sin((2π/p)*x))

// Interrupt Service Routine
void IsrOverflowSm20() {
  // Starts at 1, ends at PWM freq
  // e.g. 1 - 10,000
  // TODO: This doesn't handle fractional values for PWM
  if(vElapsedIntervals == pwmFreq) {
    // If we've reached the PWM frequency then reset the counter to avoid performance issues with large numbers
    vElapsedIntervals = 0;
  }

  float32_t updateIntervalPeriod = vUpdateIntervalPeriod;
  float32_t elapsedIntervals = ++vElapsedIntervals;
  float32_t radiansOutput;
  float32_t dutyCycle = MaxDutyCycle - 1;
  float32_t sinWave1;
  float32_t sinWave2;
  float32_t dutyCycleSinWave1Output;
  float32_t dutyCycleSinWave2Output;
  float32_t sinWave1DutyCycle;
  float32_t sinWave2DutyCycle;

  // The Teensy's LED is lit during the interrupt routine in order to be able to measure its execution time.
  digitalWriteFast (LED_BUILTIN, HIGH);

  // Example
  // Sine Frequency = 50Hz
  // PWM Frequency = 10,000Hz
  // 2π = 360°
  // Update interval = 2π(50/10000)
  // Update interval = 2π(0.005)
  // Update interval = every 0.031415926535898 rads

  // radiansOutput stores the current location along the x-axis
  arm_mult_f32(&updateIntervalPeriod, &elapsedIntervals, &radiansOutput, 1);

  // Shift so wave2 is out of phase with wave 1
  float32_t wave1Radians = radiansOutput;
  float32_t wave2Radians = radiansOutput + phaseShift;

  sinWave1 = arm_sin_f32 ( wave1Radians );
  sinWave2 = arm_sin_f32 ( wave2Radians );

  // sinWave1 & sinWave1 vary between -1 and 1, with 0 being the mid point
  arm_mult_f32(&dutyCycle, &sinWave1, &dutyCycleSinWave1Output, 1);
  arm_mult_f32(&dutyCycle, &sinWave2, &dutyCycleSinWave2Output, 1);

  // Flash LEDs at peak of wave
  // Wave 1 peak = 90°
  // Wave 2 peak = wave 1 peak + phaseShift
  // Wave 1 = Blue LED
  // Wave 2 = Red LED
  if (sinWave1 > sinWavePeak - LedTolerance && sinWave1 < sinWavePeak + LedTolerance) {
    digitalWriteFast (BlueLedPin, HIGH);
  }
  else if (sinWave1 < sinWavePeak - LedTolerance || sinWave1 > sinWavePeak + LedTolerance) {
    digitalWriteFast (BlueLedPin, LOW);
  }
  
  if (sinWave2 > sinWavePeak - LedTolerance && sinWave2 < sinWavePeak + LedTolerance) {
    digitalWriteFast (RedLedPin, HIGH);
  }
  else if (sinWave2 < sinWavePeak - LedTolerance || sinWave2 > sinWavePeak + LedTolerance) {
    digitalWriteFast (RedLedPin, LOW);
  }

  sinWave1DutyCycle = roundf ( dutyCycleSinWave1Output );
  sinWave2DutyCycle = roundf ( dutyCycleSinWave2Output );

  float32_t wave1DutyCycle = MaxDutyCycle + sinWave1DutyCycle;
  float32_t wave2DutyCycle = MaxDutyCycle + sinWave2DutyCycle;

  float32_t wave1AmplitudeOffsetAdjusted = map(wave1DutyCycle, MinDutyCycle, MaxDutyCycle * 2, 10, MaxDutyCycle * 2);
  float32_t wave2AmplitudeOffsetAdjusted = map(wave2DutyCycle, MinDutyCycle, MaxDutyCycle * 2, 10, MaxDutyCycle * 2);

  Sm20.updateDutyCycle (static_cast<uint16_t> (wave1AmplitudeOffsetAdjusted));
  Sm22.updateDutyCycle (static_cast<uint16_t> (wave2AmplitudeOffsetAdjusted));

  int resolution = 20000 / (sineFreq * 2);

  if(ShouldOutputWaveValues && vElapsedIntervals % resolution == 0) {
    // You can see these in graphical form by looking at the Arduino IDE Serial Plotter window
    Serial.print("Wave1:");
    Serial.print(sinWave1);
    Serial.print(",");
    Serial.print("Wave2:");
    Serial.print(sinWave2);
    Serial.print(",");
    Serial.print("SinWave1DutyCycle:");
    Serial.print(sinWave1DutyCycle);
    Serial.print(",");
    Serial.print("Wave1DutyCycle:");
    Serial.print(wave1DutyCycle);
    Serial.print(",");
    Serial.print("Wave1AmplitudeOffsetAdjusted:");
    Serial.print(wave1AmplitudeOffsetAdjusted);
    Serial.print(",");
    Serial.print("SinWave2DutyCycle:");
    Serial.print(sinWave2DutyCycle);
    Serial.print(",");
    Serial.print("Wave2DutyCycle:");
    Serial.print(wave2DutyCycle);
    Serial.print(",");
    Serial.print("Wave2AmplitudeOffsetAdjusted:");
    Serial.println(wave2AmplitudeOffsetAdjusted);
  }

  Sm20.clearStatusFlags (kPWM_CompareVal1Flag);
  Tm2.setPwmLdok();

  digitalWriteFast (LED_BUILTIN, LOW);
}

void preComputeValues() {
  degreesToRadiansMultiplier = PI / 180.0;
  radiansToDegreesMultiplier = 180.0 / PI;

  degrees90InRadians = 90.0 * degreesToRadiansMultiplier;
  degrees180InRadians = 180.0 * degreesToRadiansMultiplier;
  degrees270InRadians = 270.0 * degreesToRadiansMultiplier;
  degrees360InRadians = 360.0 * degreesToRadiansMultiplier;

  sinWavePeak = 1 + amplitudeShift;
  sinWaveTrough = -1 + amplitudeShift;
}

float32_t radiansToDegrees(float32_t radians) {
  return radians * radiansToDegreesMultiplier;
}

float32_t degreesToRadians(float32_t degrees) {
  return degrees * degreesToRadiansMultiplier;
}

void writeLog(String msg) {
  for (int i = 0; i < LogSize - 1; i++) {
    logs[i] = logs[i + 1];
  }

  logs[LogSize - 1] = msg;

  display.clearDisplay();

  for (int i = 0; i < LogSize; i++) {
    display.setCursor(10, (i+1)*10);
    display.println(logs[i]);
    //Serial.println(logs[i]);
  }

  display.display();  
}
